@@ -9,14 +9,17 @@ int main(int argc, char **argv)
 
   try
   {
+    uint32_t loopFrequency;
+
     ROS_INFO("SBG DRIVER - Init node, load params and connect to the device.");
     SbgDevice sbg_device(node_handle);
 
     ROS_INFO("SBG DRIVER - Initialize device for receiving data");
     sbg_device.initDeviceForReceivingData();
 
-    ROS_INFO("SBG DRIVER - START RECEIVING DATA");
-    ros::Rate loop_rate(sbg_device.getUpdateFrequency() * 2);
+    loopFrequency = sbg_device.getUpdateFrequency();
+    ROS_INFO("SBG DRIVER - ROS Node frequency : %u Hz", loopFrequency);
+    ros::Rate loop_rate(loopFrequency);
 
     while (ros::ok())
     {
@@ -32,4 +35,4 @@ int main(int argc, char **argv)
   }
 
   return 0;
-}
\ No newline at end of file
+}
